@@ -6,16 +6,23 @@ from flask import request, g
 from werkzeug.datastructures import parse_set_header
 
 from .utils import make_salt, effective_max_age, none_or_truthy
+from .recache import RECACHE_HEADER
 
-class CacheMiss(Exception):
-    pass
+class CacheMiss(Exception): pass
+class NoResourceMetadata(CacheMiss): pass
+class NoMatchingRepresentation(CacheMiss): pass
+class NotFreshEnoughForClient(CacheMiss): pass
+class RecacheRequested(CacheMiss): pass
 
 class Config(object):
     def __init__(self, resource_exemptions=(), master_salt='',
-                 request_controls_cache=True):
+                 request_controls_cache=True, preemptive_recache_seconds=0,
+                 preemptive_recache_callback=None):
         self.resource_exemptions = resource_exemptions
         self.master_salt = master_salt
         self.request_controls_cache = request_controls_cache
+        self.preemptive_recache_seconds = preemptive_recache_seconds
+        self.preemptive_recache_callback = preemptive_recache_callback
 
 class Metadata(object):
     def __init__(self, vary, salt):
@@ -40,20 +47,24 @@ class Base(object):
         return request.path
     def make_key(self, *bits):
         return self.CACHE_SEPARATOR.join(bits)
-    def metadata_cache_key(self):
-        return self.make_key('metadata', self.request_path_and_query())
-    def response_cache_key(self, metadata):
+    def make_response_key(self, namespace, metadata):
         ctx = hashlib.md5()
         for header in metadata.vary:
             ctx.update(header + request.headers.get(header, ''))
         ctx.update(metadata.salt)
         ctx.update(self.config.master_salt)
-        return self.make_key('representation', ctx.hexdigest(),
+        return self.make_key(namespace, ctx.hexdigest(),
                              self.request_path_and_query())
-    def get_or_miss(self, key, message):
+    def metadata_cache_key(self):
+        return self.make_key('metadata', self.request_path_and_query())
+    def response_cache_key(self, metadata):
+        return self.make_response_key('representation', metadata)
+    def recache_cache_key(self, metadata):
+        return self.make_response_key('recache', metadata)
+    def get_or_miss(self, key, exception):
         result = self.cache.get(key)
         if result is None:
-            raise CacheMiss(message)
+            raise exception()
         return result
     def is_exempt(self):
         for prefix in self.config.resource_exemptions:
@@ -77,13 +88,18 @@ class Retrieval(Base):
         return True
     def fetch_metadata(self):
         key = self.metadata_cache_key()
-        return self.get_or_miss(key, 'no resource metadata')
+        return self.get_or_miss(key, NoResourceMetadata)
     def fetch_response(self):
         metadata = self.fetch_metadata()
+        if request.headers.get(RECACHE_HEADER) == metadata.salt:
+            raise RecacheRequested()
         g.webcache_cache_metadata = metadata
         key = self.response_cache_key(metadata)
-        response = self.get_or_miss(key, 'no matching representation')
-        self.verify_response_freshness_or_miss(response)
+        response = self.get_or_miss(key, NoMatchingRepresentation)
+        freshness = self.response_freshness_seconds(response)
+        self.verify_response_freshness_or_miss(response, freshness)
+        if self.should_recache_preemptively(freshness, metadata):
+            self.config.preemptive_recache_callback(metadata.salt)
         g.webcache_cached_response = True
         return response
     def response_freshness_seconds(self, response):
@@ -101,16 +117,28 @@ class Retrieval(Base):
         else:
             rv = 0 # should never happen for cached responses
         return max(0, rv)
-    def verify_response_freshness_or_miss(self, response):
+    def verify_response_freshness_or_miss(self, response, freshness):
         if not self.config.request_controls_cache:
             return
         if 'min-fresh' not in request.cache_control:
             return
-        freshness = self.response_freshness_seconds(response)
         if freshness >= request.cache_control.min_fresh:
             return
-        raise CacheMiss('response not fresh enough for client')
-        
+        raise NotFreshEnoughForClient()
+    def should_recache_preemptively(self, freshness, metadata):
+        if self.config.preemptive_recache_callback is None:
+            return False
+        if self.config.preemptive_recache_seconds < freshness:
+            return False
+        key = self.recache_cache_key(metadata)
+        if self.cache.get(key):
+            return False
+        salt = make_salt()
+        self.cache.add(key, salt, self.config.preemptive_recache_seconds)
+        if self.cache.get(key) != salt:
+            return False
+        return True
+
 class Store(Base):
     def should_cache_response(self, response):
         if (response.is_streamed or        # theoretically possible
@@ -163,6 +191,9 @@ class Store(Base):
         metadata = self.get_or_create_metadata(response, expiry_seconds)
         self.mark_cache_hit(response)
         self.store_response(metadata, response, expiry_seconds)
+        self.delete_recache_key(metadata)
+    def delete_recache_key(self, metadata):
+        self.cache.delete(self.recache_cache_key(metadata))
     def mark_cache_hit(self, response):
         if self.X_CACHE_HEADER:
             response.headers[self.X_CACHE_HEADER] = 'hit'
